@@ -28,18 +28,22 @@
 #include "config.h"
 #endif
 
-#include "xf86.h"
-#include "xf86_OSproc.h"
-#include "xf86cmap.h"
+#include <xf86.h>
+#include <xf86_OSproc.h>
+#include <xf86cmap.h>
+#include <xf86drmMode.h>
+
+#include <xorgVersion.h>
+
+#if XORG_VERSION_CURRENT < XORG_VERSION_NUMERIC(1,6,99,0,0)
+#include <xf86Resources.h>
+#endif
 
 #include "common.h"
-#include "intel.h"
 #include "intel_driver.h"
 #include "legacy/legacy.h"
 #include "sna/sna_module.h"
 
-#include <xf86drmMode.h>
-
 static struct intel_device_info *chipset_info;
 
 static const struct intel_device_info intel_i81x_info = {
@@ -56,7 +60,7 @@ static const struct intel_device_info intel_i855_info = {
 	.gen = 21,
 };
 static const struct intel_device_info intel_i865_info = {
-	.gen = 21,
+	.gen = 22,
 };
 
 static const struct intel_device_info intel_i915_info = {
@@ -140,6 +144,8 @@ static const SymTabRec _intel_chipsets[] = {
 	{PCI_CHIP_IVYBRIDGE_S_GT1,		"Ivybridge Server" },
 	{-1,					NULL}
 };
+#define NUM_CHIPSETS (sizeof(_intel_chipsets) / sizeof(_intel_chipsets[0]))
+
 SymTabRec *intel_chipsets = (SymTabRec *) _intel_chipsets;
 
 #define INTEL_DEVICE_MATCH(d,i) \
@@ -274,10 +280,8 @@ static Bool has_kernel_mode_setting(struct pci_device *dev)
 		 dev->domain, dev->bus, dev->dev, dev->func);
 
 	ret = drmCheckModesettingSupported(id);
-	if (ret) {
-		if (xf86LoadKernelModule("i915"))
-			ret = drmCheckModesettingSupported(id);
-	}
+	if (ret && xf86LoadKernelModule("i915"))
+		ret = drmCheckModesettingSupported(id);
 	/* Be nice to the user and load fbcon too */
 	if (!ret)
 		(void)xf86LoadKernelModule("fbcon");
@@ -298,8 +302,8 @@ static Bool intel_pci_probe(DriverPtr		driver,
 			    intptr_t		match_data)
 {
 	ScrnInfoPtr scrn;
-	PciChipsets intel_pci_chipsets[ARRAY_SIZE(_intel_chipsets)];
-	int i;
+	PciChipsets intel_pci_chipsets[NUM_CHIPSETS];
+	unsigned i;
 
 	chipset_info = (void *)match_data;
 
@@ -319,10 +323,14 @@ static Bool intel_pci_probe(DriverPtr		driver,
 #endif
 	}
 
-	for (i = 0; i < ARRAY_SIZE(_intel_chipsets); i++) {
+	for (i = 0; i < NUM_CHIPSETS; i++) {
 		intel_pci_chipsets[i].numChipset = intel_chipsets[i].token;
 		intel_pci_chipsets[i].PCIid = intel_chipsets[i].token;
+#if XORG_VERSION_CURRENT < XORG_VERSION_NUMERIC(1,6,99,0,0)
+		intel_pci_chipsets[i].resList = RES_SHARED_VGA;
+#else
 		intel_pci_chipsets[i].dummy = NULL;
+#endif
 	}
 
 	scrn = xf86ConfigPciEntity(NULL, 0, entity_num, intel_pci_chipsets,
@@ -346,8 +354,10 @@ static Bool intel_pci_probe(DriverPtr		driver,
 		default:
 #if USE_SNA
 			sna_init_scrn(scrn, entity_num);
-#else
+#elif USE_UXA
 			intel_init_scrn(scrn);
+#else
+			scrn = NULL;
 #endif
 			break;
 		}
@@ -387,8 +397,10 @@ intel_available_options(int chipid, int busid)
 	default:
 #if USE_SNA
 		return sna_available_options(chipid, busid);
-#else
+#elif USE_UXA
 		return intel_uxa_available_options(chipid, busid);
+#else
+		return NULL;
 #endif
 	}
 }
