@@ -96,14 +96,6 @@ static void nop_done(struct sna *sna, const struct sna_composite_op *op)
 	(void)op;
 }
 
-static void blt_done(struct sna *sna, const struct sna_composite_op *op)
-{
-	struct kgem *kgem = &sna->kgem;
-
-	_kgem_set_mode(kgem, KGEM_BLT);
-	(void)op;
-}
-
 static void gen6_blt_copy_done(struct sna *sna, const struct sna_composite_op *op)
 {
 	struct kgem *kgem = &sna->kgem;
@@ -114,8 +106,6 @@ static void gen6_blt_copy_done(struct sna *sna, const struct sna_composite_op *o
 		b[1] = b[2] = 0;
 		kgem->nbatch += 3;
 	}
-
-	_kgem_set_mode(kgem, KGEM_BLT);
 	(void)op;
 }
 
@@ -127,17 +117,17 @@ static bool sna_blt_fill_init(struct sna *sna,
 			      uint32_t pixel)
 {
 	struct kgem *kgem = &sna->kgem;
-	int pitch;
 
+	assert(bo->tiling != I915_TILING_Y);
 	blt->bo[0] = bo;
 
-	pitch = bo->pitch;
+	blt->br13 = bo->pitch;
 	blt->cmd = XY_SCANLINE_BLT;
-	if (kgem->gen >= 40 && blt->bo[0]->tiling) {
-		blt->cmd |= 1 << 11;
-		pitch >>= 2;
+	if (kgem->gen >= 40 && bo->tiling) {
+		blt->cmd |= BLT_DST_TILED;
+		blt->br13 >>= 2;
 	}
-	assert(pitch < MAXSHORT);
+	assert(blt->br13 < MAXSHORT);
 
 	if (alu == GXclear)
 		pixel = 0;
@@ -148,7 +138,7 @@ static bool sna_blt_fill_init(struct sna *sna,
 			alu = GXset;
 	}
 
-	blt->br13 = 1<<31 | (fill_ROP[alu] << 16) | pitch;
+	blt->br13 |= 1<<31 | (fill_ROP[alu] << 16);
 	switch (bpp) {
 	default: assert(0);
 	case 32: blt->br13 |= 1 << 25; /* RGB8888 */
@@ -163,10 +153,11 @@ static bool sna_blt_fill_init(struct sna *sna,
 	if (!kgem_check_bo_fenced(kgem, bo, NULL) ||
 	    !kgem_check_batch(kgem, 12)) {
 		_kgem_submit(kgem);
+		assert(kgem_check_bo_fenced(kgem, bo, NULL));
 		_kgem_set_mode(kgem, KGEM_BLT);
 	}
 
-	if (sna->blt_state.fill_bo != bo->handle ||
+	if (sna->blt_state.fill_bo != bo->unique_id ||
 	    sna->blt_state.fill_pixel != pixel ||
 	    sna->blt_state.fill_alu != alu)
 	{
@@ -195,7 +186,7 @@ static bool sna_blt_fill_init(struct sna *sna,
 		b[8] = 0;
 		kgem->nbatch += 9;
 
-		sna->blt_state.fill_bo = bo->handle;
+		sna->blt_state.fill_bo = bo->unique_id;
 		sna->blt_state.fill_pixel = pixel;
 		sna->blt_state.fill_alu = alu;
 	}
@@ -245,7 +236,7 @@ inline static void sna_blt_fill_one(struct sna *sna,
 
 	assert(x >= 0);
 	assert(y >= 0);
-	assert((y+height) * blt->bo[0]->pitch <= blt->bo[0]->size);
+	assert((y+height) * blt->bo[0]->pitch <= kgem_bo_size(blt->bo[0]));
 
 	if (!kgem_check_batch(kgem, 3))
 		sna_blt_fill_begin(sna, blt);
@@ -300,6 +291,8 @@ static Bool sna_blt_copy_init(struct sna *sna,
 	kgem_set_mode(kgem, KGEM_BLT);
 	if (!kgem_check_bo_fenced(kgem, src, dst, NULL)) {
 		_kgem_submit(kgem);
+		if (!kgem_check_bo_fenced(kgem, src, dst, NULL))
+			return FALSE;
 		_kgem_set_mode(kgem, KGEM_BLT);
 	}
 
@@ -307,6 +300,106 @@ static Bool sna_blt_copy_init(struct sna *sna,
 	return TRUE;
 }
 
+static Bool sna_blt_alpha_fixup_init(struct sna *sna,
+				     struct sna_blt_state *blt,
+				     struct kgem_bo *src,
+				     struct kgem_bo *dst,
+				     int bpp, uint32_t alpha)
+{
+	struct kgem *kgem = &sna->kgem;
+
+	blt->bo[0] = src;
+	blt->bo[1] = dst;
+
+	blt->cmd = XY_FULL_MONO_PATTERN_BLT;
+	blt->pitch[0] = src->pitch;
+	if (kgem->gen >= 40 && src->tiling) {
+		blt->cmd |= BLT_SRC_TILED;
+		blt->pitch[0] >>= 2;
+	}
+	assert(blt->pitch[0] < MAXSHORT);
+
+	blt->pitch[1] = dst->pitch;
+	if (kgem->gen >= 40 && dst->tiling) {
+		blt->cmd |= BLT_DST_TILED;
+		blt->pitch[1] >>= 2;
+	}
+	assert(blt->pitch[1] < MAXSHORT);
+
+	blt->overwrites = 1;
+	blt->br13 = (0xfc << 16) | blt->pitch[1];
+	switch (bpp) {
+	default: assert(0);
+	case 32: blt->cmd |= BLT_WRITE_ALPHA | BLT_WRITE_RGB;
+		 blt->br13 |= 1 << 25; /* RGB8888 */
+	case 16: blt->br13 |= 1 << 24; /* RGB565 */
+	case 8: break;
+	}
+	blt->pixel = alpha;
+
+	kgem_set_mode(kgem, KGEM_BLT);
+	if (!kgem_check_bo_fenced(kgem, src, dst, NULL)) {
+		_kgem_submit(kgem);
+		if (!kgem_check_bo_fenced(kgem, src, dst, NULL))
+			return FALSE;
+		_kgem_set_mode(kgem, KGEM_BLT);
+	}
+
+	sna->blt_state.fill_bo = 0;
+	return TRUE;
+}
+
+static void sna_blt_alpha_fixup_one(struct sna *sna,
+				    const struct sna_blt_state *blt,
+				    int src_x, int src_y,
+				    int width, int height,
+				    int dst_x, int dst_y)
+{
+	struct kgem *kgem = &sna->kgem;
+	uint32_t *b;
+
+	DBG(("%s: (%d, %d) -> (%d, %d) x (%d, %d)\n",
+	     __FUNCTION__, src_x, src_y, dst_x, dst_y, width, height));
+
+	assert(src_x >= 0);
+	assert(src_y >= 0);
+	assert((src_y + height) * blt->bo[0]->pitch <= kgem_bo_size(blt->bo[0]));
+	assert(dst_x >= 0);
+	assert(dst_y >= 0);
+	assert((dst_y + height) * blt->bo[1]->pitch <= kgem_bo_size(blt->bo[1]));
+	assert(width > 0);
+	assert(height > 0);
+
+	if (!kgem_check_batch(kgem, 12) || !kgem_check_reloc(kgem, 2)) {
+		_kgem_submit(kgem);
+		_kgem_set_mode(kgem, KGEM_BLT);
+	}
+
+	b = kgem->batch + kgem->nbatch;
+	b[0] = blt->cmd;
+	b[1] = blt->br13;
+	b[2] = (dst_y << 16) | dst_x;
+	b[3] = ((dst_y + height) << 16) | (dst_x + width);
+	b[4] = kgem_add_reloc(kgem, kgem->nbatch + 4,
+			      blt->bo[1],
+			      I915_GEM_DOMAIN_RENDER << 16 |
+			      I915_GEM_DOMAIN_RENDER |
+			      KGEM_RELOC_FENCED,
+			      0);
+	b[5] = blt->pitch[0];
+	b[6] = (src_y << 16) | src_x;
+	b[7] = kgem_add_reloc(kgem, kgem->nbatch + 7,
+			      blt->bo[0],
+			      I915_GEM_DOMAIN_RENDER << 16 |
+			      KGEM_RELOC_FENCED,
+			      0);
+	b[8] = blt->pixel;
+	b[9] = blt->pixel;
+	b[10] = 0;
+	b[11] = 0;
+	kgem->nbatch += 12;
+}
+
 static void sna_blt_copy_one(struct sna *sna,
 			     const struct sna_blt_state *blt,
 			     int src_x, int src_y,
@@ -321,10 +414,10 @@ static void sna_blt_copy_one(struct sna *sna,
 
 	assert(src_x >= 0);
 	assert(src_y >= 0);
-	assert((src_y + height) * blt->bo[0]->pitch <= blt->bo[0]->size);
+	assert((src_y + height) * blt->bo[0]->pitch <= kgem_bo_size(blt->bo[0]));
 	assert(dst_x >= 0);
 	assert(dst_y >= 0);
-	assert((dst_y + height) * blt->bo[1]->pitch <= blt->bo[1]->size);
+	assert((dst_y + height) * blt->bo[1]->pitch <= kgem_bo_size(blt->bo[1]));
 	assert(width > 0);
 	assert(height > 0);
 
@@ -455,7 +548,7 @@ get_rgba_from_pixel(uint32_t pixel,
 }
 
 Bool
-sna_get_pixel_from_rgba(uint32_t * pixel,
+_sna_get_pixel_from_rgba(uint32_t * pixel,
 			uint16_t red,
 			uint16_t green,
 			uint16_t blue,
@@ -465,20 +558,6 @@ sna_get_pixel_from_rgba(uint32_t * pixel,
 	int rbits, bbits, gbits, abits;
 	int rshift, bshift, gshift, ashift;
 
-	switch (format) {
-	case PICT_x8r8g8b8:
-		alpha = 0xffff;
-	case PICT_a8r8g8b8:
-		*pixel = ((alpha >> 8 << 24) |
-			  (red >> 8 << 16) |
-			  (green & 0xff00) |
-			  (blue >> 8));
-		return TRUE;
-	case PICT_a8:
-		*pixel = alpha >> 8;
-		return TRUE;
-	}
-
 	rbits = PICT_FORMAT_R(format);
 	gbits = PICT_FORMAT_G(format);
 	bbits = PICT_FORMAT_B(format);
@@ -559,7 +638,9 @@ get_pixel(PicturePtr picture)
 
 	DBG(("%s: %p\n", __FUNCTION__, pixmap));
 
-	sna_pixmap_move_to_cpu(pixmap, false);
+	if (!sna_pixmap_move_to_cpu(pixmap, MOVE_READ))
+		return 0;
+
 	switch (pixmap->drawable.bitsPerPixel) {
 	case 32: return *(uint32_t *)pixmap->devPrivate.ptr;
 	case 16: return *(uint16_t *)pixmap->devPrivate.ptr;
@@ -601,7 +682,8 @@ sna_picture_is_solid(PicturePtr picture, uint32_t *color)
 	if (!is_solid(picture))
 		return FALSE;
 
-	*color = get_solid_color(picture, PICT_a8r8g8b8);
+	if (color)
+		*color = get_solid_color(picture, PICT_a8r8g8b8);
 	return TRUE;
 }
 
@@ -711,7 +793,7 @@ inline static void _sna_blt_fill_box(struct sna *sna,
 
 	assert(box->x1 >= 0);
 	assert(box->y1 >= 0);
-	assert(box->y2 * blt->bo[0]->pitch <= blt->bo[0]->size);
+	assert(box->y2 * blt->bo[0]->pitch <= kgem_bo_size(blt->bo[0]));
 
 	if (!kgem_check_batch(kgem, 3))
 		sna_blt_fill_begin(sna, blt);
@@ -833,7 +915,7 @@ prepare_blt_clear(struct sna *sna,
 		op->box   = blt_composite_fill_box_no_offset;
 		op->boxes = blt_composite_fill_boxes_no_offset;
 	}
-	op->done  = blt_done;
+	op->done  = nop_done;
 
 	return sna_blt_fill_init(sna, &op->u.blt,
 				 op->dst.bo,
@@ -856,7 +938,7 @@ prepare_blt_fill(struct sna *sna,
 		op->box   = blt_composite_fill_box_no_offset;
 		op->boxes = blt_composite_fill_boxes_no_offset;
 	}
-	op->done  = blt_done;
+	op->done  = nop_done;
 
 	return sna_blt_fill_init(sna, &op->u.blt, op->dst.bo,
 				 op->dst.pixmap->drawable.bitsPerPixel,
@@ -942,48 +1024,141 @@ static void blt_composite_copy_boxes(struct sna *sna,
 	} while(--nbox);
 }
 
+fastcall static void
+blt_composite_copy_with_alpha(struct sna *sna,
+			      const struct sna_composite_op *op,
+			      const struct sna_composite_rectangles *r)
+{
+	int x1, x2, y1, y2;
+	int src_x, src_y;
+
+	DBG(("%s: src=(%d, %d), dst=(%d, %d), size=(%d, %d)\n",
+	     __FUNCTION__,
+	     r->src.x, r->src.y,
+	     r->dst.x, r->dst.y,
+	     r->width, r->height));
+
+	/* XXX higher layer should have clipped? */
+
+	x1 = r->dst.x + op->dst.x;
+	y1 = r->dst.y + op->dst.y;
+	x2 = x1 + r->width;
+	y2 = y1 + r->height;
+
+	src_x = r->src.x - x1;
+	src_y = r->src.y - y1;
+
+	/* clip against dst */
+	if (x1 < 0)
+		x1 = 0;
+	if (y1 < 0)
+		y1 = 0;
+
+	if (x2 > op->dst.width)
+		x2 = op->dst.width;
+
+	if (y2 > op->dst.height)
+		y2 = op->dst.height;
+
+	DBG(("%s: box=(%d, %d), (%d, %d)\n", __FUNCTION__, x1, y1, x2, y2));
+
+	if (x2 <= x1 || y2 <= y1)
+		return;
+
+	sna_blt_alpha_fixup_one(sna, &op->u.blt,
+				x1 + src_x, y1 + src_y,
+				x2 - x1, y2 - y1,
+				x1, y1);
+}
+
+fastcall static void
+blt_composite_copy_box_with_alpha(struct sna *sna,
+				  const struct sna_composite_op *op,
+				  const BoxRec *box)
+{
+	DBG(("%s: box (%d, %d), (%d, %d)\n",
+	     __FUNCTION__, box->x1, box->y1, box->x2, box->y2));
+	sna_blt_alpha_fixup_one(sna, &op->u.blt,
+				box->x1 + op->u.blt.sx,
+				box->y1 + op->u.blt.sy,
+				box->x2 - box->x1,
+				box->y2 - box->y1,
+				box->x1 + op->dst.x,
+				box->y1 + op->dst.y);
+}
+
+static void
+blt_composite_copy_boxes_with_alpha(struct sna *sna,
+				    const struct sna_composite_op *op,
+				    const BoxRec *box, int nbox)
+{
+	DBG(("%s: nbox=%d\n", __FUNCTION__, nbox));
+	do {
+		DBG(("%s: box (%d, %d), (%d, %d)\n",
+		     __FUNCTION__, box->x1, box->y1, box->x2, box->y2));
+		sna_blt_alpha_fixup_one(sna, &op->u.blt,
+					box->x1 + op->u.blt.sx, box->y1 + op->u.blt.sy,
+					box->x2 - box->x1, box->y2 - box->y1,
+					box->x1 + op->dst.x, box->y1 + op->dst.y);
+		box++;
+	} while(--nbox);
+}
+
 static Bool
 prepare_blt_copy(struct sna *sna,
-		 struct sna_composite_op *op)
+		 struct sna_composite_op *op,
+		 uint32_t alpha_fixup)
 {
 	PixmapPtr src = op->u.blt.src_pixmap;
 	struct sna_pixmap *priv = sna_pixmap(src);
 
-	if (priv->gpu_bo->tiling == I915_TILING_Y)
+	if (!kgem_bo_can_blt(&sna->kgem, priv->gpu_bo))
 		return FALSE;
 
-	if (!kgem_check_bo_fenced(&sna->kgem, priv->gpu_bo, NULL)) {
+	if (!kgem_check_bo_fenced(&sna->kgem, op->dst.bo, priv->gpu_bo, NULL)) {
 		_kgem_submit(&sna->kgem);
+		if (!kgem_check_bo_fenced(&sna->kgem,
+					  op->dst.bo, priv->gpu_bo, NULL))
+			return FALSE;
 		_kgem_set_mode(&sna->kgem, KGEM_BLT);
 	}
 
 	DBG(("%s\n", __FUNCTION__));
 
-	op->blt   = blt_composite_copy;
-	op->box   = blt_composite_copy_box;
-	op->boxes = blt_composite_copy_boxes;
 	if (sna->kgem.gen >= 60)
 		op->done  = gen6_blt_copy_done;
 	else
-		op->done  = blt_done;
+		op->done  = nop_done;
 
-	return sna_blt_copy_init(sna, &op->u.blt,
-				 priv->gpu_bo,
-				 op->dst.bo,
-				 src->drawable.bitsPerPixel,
-				 GXcopy);
+	if (alpha_fixup) {
+		op->blt   = blt_composite_copy_with_alpha;
+		op->box   = blt_composite_copy_box_with_alpha;
+		op->boxes = blt_composite_copy_boxes_with_alpha;
+
+		return sna_blt_alpha_fixup_init(sna, &op->u.blt,
+						priv->gpu_bo,
+						op->dst.bo,
+						src->drawable.bitsPerPixel,
+						alpha_fixup);
+	} else {
+		op->blt   = blt_composite_copy;
+		op->box   = blt_composite_copy_box;
+		op->boxes = blt_composite_copy_boxes;
+
+		return sna_blt_copy_init(sna, &op->u.blt,
+					 priv->gpu_bo,
+					 op->dst.bo,
+					 src->drawable.bitsPerPixel,
+					 GXcopy);
+	}
 }
 
 static void blt_vmap_done(struct sna *sna, const struct sna_composite_op *op)
 {
 	struct kgem_bo *bo = (struct kgem_bo *)op->u.blt.src_pixmap;
 
-	blt_done(sna, op);
-	if (bo) {
-		struct kgem *kgem = &sna->kgem;
-		kgem_bo_sync(kgem, bo, true);
-		kgem_bo_destroy(kgem, bo);
-	}
+	if (bo)
+		kgem_bo_destroy(&sna->kgem, bo);
 }
 
 fastcall static void
@@ -1010,10 +1185,8 @@ blt_put_composite(struct sna *sna,
 		data += (src_x - dst_x) * bpp / 8;
 		data += (src_y - dst_y) * pitch;
 
-		dst_priv->gpu_bo =
-			sna_replace(sna, dst_priv->gpu_bo,
-				    r->width, r->height, bpp,
-				    data, pitch);
+		sna_replace(sna, op->dst.pixmap, &dst_priv->gpu_bo,
+			    data, pitch);
 	} else {
 		BoxRec box;
 
@@ -1022,9 +1195,9 @@ blt_put_composite(struct sna *sna,
 		box.x2 = dst_x + r->width;
 		box.y2 = dst_y + r->height;
 
-		sna_write_boxes(sna,
+		sna_write_boxes(sna, dst,
 				dst_priv->gpu_bo, 0, 0,
-				data, pitch, bpp, src_x, src_y,
+				data, pitch, src_x, src_y,
 				&box, 1);
 	}
 }
@@ -1050,19 +1223,13 @@ fastcall static void blt_put_composite_box(struct sna *sna,
 		data += (box->y1 + op->u.blt.sy) * pitch;
 		data += (box->x1 + op->u.blt.sx) * bpp;
 
-		dst_priv->gpu_bo =
-			sna_replace(sna,
-				    op->dst.bo,
-				    op->dst.width,
-				    op->dst.height,
-				    src->drawable.bitsPerPixel,
-				    data, pitch);
+		sna_replace(sna, op->dst.pixmap, &dst_priv->gpu_bo,
+			    data, pitch);
 	} else {
-		sna_write_boxes(sna,
+		sna_write_boxes(sna, op->dst.pixmap,
 				op->dst.bo, op->dst.x, op->dst.y,
 				src->devPrivate.ptr,
 				src->devKind,
-				src->drawable.bitsPerPixel,
 				op->u.blt.sx, op->u.blt.sy,
 				box, 1);
 	}
@@ -1090,41 +1257,152 @@ static void blt_put_composite_boxes(struct sna *sna,
 		data += (box->y1 + op->u.blt.sy) * pitch;
 		data += (box->x1 + op->u.blt.sx) * bpp;
 
-		dst_priv->gpu_bo =
-			sna_replace(sna,
-				    op->dst.bo,
-				    op->dst.width,
-				    op->dst.height,
-				    src->drawable.bitsPerPixel,
-				    data, pitch);
+		sna_replace(sna, op->dst.pixmap, &dst_priv->gpu_bo,
+			    data, pitch);
 	} else {
-		sna_write_boxes(sna,
+		sna_write_boxes(sna, op->dst.pixmap,
 				op->dst.bo, op->dst.x, op->dst.y,
 				src->devPrivate.ptr,
 				src->devKind,
-				src->drawable.bitsPerPixel,
 				op->u.blt.sx, op->u.blt.sy,
 				box, n);
 	}
 }
 
-static Bool
-prepare_blt_put(struct sna *sna,
-		struct sna_composite_op *op)
+fastcall static void
+blt_put_composite_with_alpha(struct sna *sna,
+			     const struct sna_composite_op *op,
+			     const struct sna_composite_rectangles *r)
+{
+	PixmapPtr dst = op->dst.pixmap;
+	PixmapPtr src = op->u.blt.src_pixmap;
+	struct sna_pixmap *dst_priv = sna_pixmap(dst);
+	int pitch = src->devKind;
+	char *data = src->devPrivate.ptr;
+
+	int16_t dst_x = r->dst.x + op->dst.x;
+	int16_t dst_y = r->dst.y + op->dst.y;
+	int16_t src_x = r->src.x + op->u.blt.sx;
+	int16_t src_y = r->src.y + op->u.blt.sy;
+
+	if (!dst_priv->pinned &&
+	    dst_x <= 0 && dst_y <= 0 &&
+	    dst_x + r->width >= op->dst.width &&
+	    dst_y + r->height >= op->dst.height) {
+		int bpp = dst->drawable.bitsPerPixel / 8;
+
+		data += (src_x - dst_x) * bpp;
+		data += (src_y - dst_y) * pitch;
+
+		dst_priv->gpu_bo =
+			sna_replace__xor(sna, op->dst.pixmap, dst_priv->gpu_bo,
+					 data, pitch,
+					 0xffffffff, op->u.blt.pixel);
+	} else {
+		BoxRec box;
+
+		box.x1 = dst_x;
+		box.y1 = dst_y;
+		box.x2 = dst_x + r->width;
+		box.y2 = dst_y + r->height;
+
+		sna_write_boxes__xor(sna, dst,
+				     dst_priv->gpu_bo, 0, 0,
+				     data, pitch, src_x, src_y,
+				     &box, 1,
+				     0xffffffff, op->u.blt.pixel);
+	}
+}
+
+fastcall static void
+blt_put_composite_box_with_alpha(struct sna *sna,
+				 const struct sna_composite_op *op,
+				 const BoxRec *box)
 {
 	PixmapPtr src = op->u.blt.src_pixmap;
-	struct sna_pixmap *priv = sna_pixmap(src);
+	struct sna_pixmap *dst_priv = sna_pixmap(op->dst.pixmap);
+
+	DBG(("%s: src=(%d, %d), dst=(%d, %d)\n", __FUNCTION__,
+	     op->u.blt.sx, op->u.blt.sy,
+	     op->dst.x, op->dst.y));
+
+	if (!dst_priv->pinned &&
+	    box->x2 - box->x1 == op->dst.width &&
+	    box->y2 - box->y1 == op->dst.height) {
+		int pitch = src->devKind;
+		int bpp = src->drawable.bitsPerPixel / 8;
+		char *data = src->devPrivate.ptr;
+
+		data += (box->y1 + op->u.blt.sy) * pitch;
+		data += (box->x1 + op->u.blt.sx) * bpp;
+
+		dst_priv->gpu_bo =
+			sna_replace__xor(sna, op->dst.pixmap, op->dst.bo,
+					 data, pitch,
+					 0xffffffff, op->u.blt.pixel);
+	} else {
+		sna_write_boxes__xor(sna, op->dst.pixmap,
+				     op->dst.bo, op->dst.x, op->dst.y,
+				     src->devPrivate.ptr,
+				     src->devKind,
+				     op->u.blt.sx, op->u.blt.sy,
+				     box, 1,
+				     0xffffffff, op->u.blt.pixel);
+	}
+}
+
+static void
+blt_put_composite_boxes_with_alpha(struct sna *sna,
+				   const struct sna_composite_op *op,
+				   const BoxRec *box, int n)
+{
+	PixmapPtr src = op->u.blt.src_pixmap;
+	struct sna_pixmap *dst_priv = sna_pixmap(op->dst.pixmap);
+
+	DBG(("%s: src=(%d, %d), dst=(%d, %d), [(%d, %d), (%d, %d) x %d]\n", __FUNCTION__,
+	     op->u.blt.sx, op->u.blt.sy,
+	     op->dst.x, op->dst.y,
+	     box->x1, box->y1, box->x2, box->y2, n));
+
+	if (n == 1 && !dst_priv->pinned &&
+	    box->x2 - box->x1 == op->dst.width &&
+	    box->y2 - box->y1 == op->dst.height) {
+		int pitch = src->devKind;
+		int bpp = src->drawable.bitsPerPixel / 8;
+		char *data = src->devPrivate.ptr;
+
+		data += (box->y1 + op->u.blt.sy) * pitch;
+		data += (box->x1 + op->u.blt.sx) * bpp;
+
+		dst_priv->gpu_bo =
+			sna_replace__xor(sna, op->dst.pixmap, op->dst.bo,
+					 data, pitch,
+					 0xffffffff, op->u.blt.pixel);
+	} else {
+		sna_write_boxes__xor(sna, op->dst.pixmap,
+				     op->dst.bo, op->dst.x, op->dst.y,
+				     src->devPrivate.ptr,
+				     src->devKind,
+				     op->u.blt.sx, op->u.blt.sy,
+				     box, n,
+				     0xffffffff, op->u.blt.pixel);
+	}
+}
+
+static Bool
+prepare_blt_put(struct sna *sna,
+		struct sna_composite_op *op,
+		uint32_t alpha_fixup)
+{
+	PixmapPtr src = op->u.blt.src_pixmap;
+	struct sna_pixmap *priv = sna_pixmap_attach(src);
 	struct kgem_bo *src_bo = NULL;
 	struct kgem_bo *free_bo = NULL;
 
 	DBG(("%s\n", __FUNCTION__));
 
 	if (priv) {
-		if (!priv->gpu_only) {
-			src_bo = priv->cpu_bo;
-			if (!src_bo)
-				src_bo = pixmap_vmap(&sna->kgem, src);
-		}
+		src_bo = priv->cpu_bo;
 	} else {
 		src_bo = kgem_create_map(&sna->kgem,
 					 src->devPrivate.ptr,
@@ -1133,23 +1411,46 @@ prepare_blt_put(struct sna *sna,
 		free_bo = src_bo;
 	}
 	if (src_bo) {
-		op->blt   = blt_composite_copy;
-		op->box   = blt_composite_copy_box;
-		op->boxes = blt_composite_copy_boxes;
-
 		op->u.blt.src_pixmap = (void *)free_bo;
 		op->done = blt_vmap_done;
 
 		src_bo->pitch = src->devKind;
-		if (!sna_blt_copy_init(sna, &op->u.blt,
-				       src_bo, op->dst.bo,
-				       op->dst.pixmap->drawable.bitsPerPixel,
-				       GXcopy))
-			return FALSE;
+		if (alpha_fixup) {
+			op->blt   = blt_composite_copy_with_alpha;
+			op->box   = blt_composite_copy_box_with_alpha;
+			op->boxes = blt_composite_copy_boxes_with_alpha;
+
+			return sna_blt_alpha_fixup_init(sna, &op->u.blt,
+							src_bo, op->dst.bo,
+							op->dst.pixmap->drawable.bitsPerPixel,
+							alpha_fixup);
+		} else {
+			op->blt   = blt_composite_copy;
+			op->box   = blt_composite_copy_box;
+			op->boxes = blt_composite_copy_boxes;
+
+			return sna_blt_copy_init(sna, &op->u.blt,
+						 src_bo, op->dst.bo,
+						 op->dst.pixmap->drawable.bitsPerPixel,
+						 GXcopy);
+		}
 	} else {
-		op->blt   = blt_put_composite;
-		op->box   = blt_put_composite_box;
-		op->boxes = blt_put_composite_boxes;
+		if (alpha_fixup)
+			return FALSE; /* XXX */
+
+		if (!sna_pixmap_move_to_cpu(src, MOVE_READ))
+			return FALSE;
+
+		if (alpha_fixup) {
+			op->u.blt.pixel = alpha_fixup;
+			op->blt   = blt_put_composite_with_alpha;
+			op->box   = blt_put_composite_box_with_alpha;
+			op->boxes = blt_put_composite_boxes_with_alpha;
+		} else {
+			op->blt   = blt_put_composite;
+			op->box   = blt_put_composite_box;
+			op->boxes = blt_put_composite_boxes;
+		}
 		op->done  = nop_done;
 	}
 
@@ -1169,11 +1470,16 @@ has_gpu_area(PixmapPtr pixmap, int x, int y, int w, int h)
 
 	if (priv->cpu_damage == NULL)
 		return TRUE;
+	if (priv->cpu_damage->mode == DAMAGE_ALL)
+		return FALSE;
 
 	area.x1 = x;
 	area.y1 = y;
 	area.x2 = x + w;
 	area.y2 = y + h;
+	if (priv->gpu_damage &&
+	    sna_damage_contains_box__no_reduce(priv->gpu_damage, &area))
+		return TRUE;
 
 	return sna_damage_contains_box(priv->cpu_damage,
 				       &area) == PIXMAN_REGION_OUT;
@@ -1187,18 +1493,19 @@ has_cpu_area(PixmapPtr pixmap, int x, int y, int w, int h)
 
 	if (!priv)
 		return TRUE;
-	if (!priv->gpu_bo)
-		return TRUE;
-	if (priv->gpu_only)
-		return FALSE;
-
 	if (priv->gpu_damage == NULL)
 		return TRUE;
+	if (priv->gpu_damage->mode == DAMAGE_ALL)
+		return FALSE;
 
 	area.x1 = x;
 	area.y1 = y;
 	area.x2 = x + w;
 	area.y2 = y + h;
+	if (priv->cpu_damage &&
+	    sna_damage_contains_box__no_reduce(priv->cpu_damage, &area))
+		return TRUE;
+
 	return sna_damage_contains_box(priv->gpu_damage,
 				       &area) == PIXMAN_REGION_OUT;
 }
@@ -1210,19 +1517,31 @@ reduce_damage(struct sna_composite_op *op,
 {
 	BoxRec r;
 
-	if (op->damage == NULL)
+	if (op->damage == NULL || *op->damage == NULL)
 		return;
 
+	if ((*op->damage)->mode == DAMAGE_ALL) {
+		op->damage = NULL;
+		return;
+	}
+
 	r.x1 = dst_x + op->dst.x;
 	r.x2 = r.x1 + width;
 
 	r.y1 = dst_y + op->dst.y;
 	r.y2 = r.y1 + height;
 
-	if (sna_damage_contains_box(*op->damage, &r) == PIXMAN_REGION_IN)
+	if (sna_damage_contains_box__no_reduce(*op->damage, &r))
 		op->damage = NULL;
 }
 
+#define alphaless(format) PICT_FORMAT(PICT_FORMAT_BPP(format),		\
+				      PICT_FORMAT_TYPE(format),		\
+				      0,				\
+				      PICT_FORMAT_R(format),		\
+				      PICT_FORMAT_G(format),		\
+				      PICT_FORMAT_B(format))
+
 Bool
 sna_blt_composite(struct sna *sna,
 		  uint32_t op,
@@ -1233,10 +1552,11 @@ sna_blt_composite(struct sna *sna,
 		  int16_t width, int16_t height,
 		  struct sna_composite_op *tmp)
 {
-	struct sna_blt_state *blt = &tmp->u.blt;
 	PictFormat src_format = src->format;
+	PixmapPtr src_pixmap;
 	struct sna_pixmap *priv;
 	int16_t tx, ty;
+	uint32_t alpha_fixup;
 	Bool ret;
 
 #if DEBUG_NO_BLT || NO_BLT_COMPOSITE
@@ -1258,10 +1578,14 @@ sna_blt_composite(struct sna *sna,
 	}
 
 	tmp->dst.pixmap = get_drawable_pixmap(dst->pDrawable);
-	priv = sna_pixmap_move_to_gpu(tmp->dst.pixmap);
-	if (priv == NULL || priv->gpu_bo->tiling == I915_TILING_Y) {
-		DBG(("%s: dst not on the gpu or using Y-tiling\n",
-		     __FUNCTION__));
+	priv = sna_pixmap_move_to_gpu(tmp->dst.pixmap, MOVE_WRITE | MOVE_READ);
+	if (priv == NULL) {
+		DBG(("%s: dst not attached\n", __FUNCTION__));
+		return FALSE;
+	}
+	if (!kgem_bo_can_blt(&sna->kgem, priv->gpu_bo)) {
+		DBG(("%s: can not blit to dst, tiling? %d, pitch? %d\n",
+		     __FUNCTION__, priv->gpu_bo->tiling, priv->gpu_bo->pitch));
 		return FALSE;
 	}
 
@@ -1279,6 +1603,7 @@ sna_blt_composite(struct sna *sna,
 
 	if (!kgem_check_bo_fenced(&sna->kgem, priv->gpu_bo, NULL)) {
 		_kgem_submit(&sna->kgem);
+		assert(kgem_check_bo_fenced(&sna->kgem, priv->gpu_bo, NULL));
 		_kgem_set_mode(&sna->kgem, KGEM_BLT);
 	}
 
@@ -1323,18 +1648,6 @@ sna_blt_composite(struct sna *sna,
 		return FALSE;
 	}
 
-	if (!(dst->format == src_format ||
-	      dst->format == PICT_FORMAT(PICT_FORMAT_BPP(src_format),
-					 PICT_FORMAT_TYPE(src_format),
-					 0,
-					 PICT_FORMAT_R(src_format),
-					 PICT_FORMAT_G(src_format),
-					 PICT_FORMAT_B(src_format)))) {
-		DBG(("%s: incompatible src/dst formats src=%08x, dst=%08x\n",
-		     __FUNCTION__, (unsigned)src_format, dst->format));
-		return FALSE;
-	}
-
 	if (!sna_transform_is_integer_translation(src->transform, &tx, &ty)) {
 		DBG(("%s: source transform is not an integer translation\n",
 		     __FUNCTION__));
@@ -1343,38 +1656,66 @@ sna_blt_composite(struct sna *sna,
 	x += tx;
 	y += ty;
 
-	/* XXX tiling? */
-	if (x < 0 || y < 0 ||
-	    x + width > src->pDrawable->width ||
-	    y + height > src->pDrawable->height) {
-		DBG(("%s: source extends outside of valid area\n",
+	if ((x > src->pDrawable->width ||
+	     y > src->pDrawable->height ||
+	     x + width < 0 ||
+	     y + height < 0) &&
+	    (!src->repeat || src->repeatType == RepeatNone)) {
+		DBG(("%s: source is outside of valid area, converting to clear\n",
 		     __FUNCTION__));
+		return prepare_blt_clear(sna, tmp);
+	}
+
+	alpha_fixup = 0;
+	if (!(dst->format == src_format ||
+	      dst->format == alphaless(src_format) ||
+	      (alphaless(dst->format) == alphaless(src_format) &&
+	       sna_get_pixel_from_rgba(&alpha_fixup,
+				       0, 0, 0, 0xffff,
+				       dst->format)))) {
+		DBG(("%s: incompatible src/dst formats src=%08x, dst=%08x\n",
+		     __FUNCTION__, (unsigned)src_format, dst->format));
 		return FALSE;
 	}
 
-	blt->src_pixmap = get_drawable_pixmap(src->pDrawable);
-	get_drawable_deltas(src->pDrawable, blt->src_pixmap, &tx, &ty);
+	/* XXX tiling? fixup extend none? */
+	if (x < 0 || y < 0 ||
+	    x + width > src->pDrawable->width ||
+	    y + height > src->pDrawable->height) {
+		DBG(("%s: source extends outside (%d, %d), (%d, %d) of valid drawable %dx%d\n",
+		     __FUNCTION__,
+		     x, y, x+width, y+width, src->pDrawable->width, src->pDrawable->height));
+		return FALSE;
+	}
+
+	src_pixmap = get_drawable_pixmap(src->pDrawable);
+	get_drawable_deltas(src->pDrawable, src_pixmap, &tx, &ty);
 	x += tx + src->pDrawable->x;
 	y += ty + src->pDrawable->y;
-	assert(x >= 0);
-	assert(y >= 0);
-	assert(x + width <= blt->src_pixmap->drawable.width);
-	assert(y + height <= blt->src_pixmap->drawable.height);
+	if (x < 0 || y < 0 ||
+	    x + width  > src_pixmap->drawable.width ||
+	    y + height > src_pixmap->drawable.height) {
+		DBG(("%s: source extends outside (%d, %d), (%d, %d) of valid pixmap %dx%d\n",
+		     __FUNCTION__,
+		     x, y, x+width, y+width, src_pixmap->drawable.width, src_pixmap->drawable.height));
+		return FALSE;
+	}
 
+	tmp->u.blt.src_pixmap = src_pixmap;
 	tmp->u.blt.sx = x - dst_x;
 	tmp->u.blt.sy = y - dst_y;
-	DBG(("%s: blt dst offset (%d, %d), source offset (%d, %d)\n",
+	DBG(("%s: blt dst offset (%d, %d), source offset (%d, %d), with alpha fixup? %x\n",
 	     __FUNCTION__,
-	     tmp->dst.x, tmp->dst.y, tmp->u.blt.sx, tmp->u.blt.sy));
+	     tmp->dst.x, tmp->dst.y, tmp->u.blt.sx, tmp->u.blt.sy, alpha_fixup));
 
-	if (has_gpu_area(blt->src_pixmap, x, y, width, height))
-		ret = prepare_blt_copy(sna, tmp);
-	else if (has_cpu_area(blt->src_pixmap, x, y, width, height))
-		ret = prepare_blt_put(sna, tmp);
-	else if (sna_pixmap_move_to_gpu(blt->src_pixmap))
-		ret = prepare_blt_copy(sna, tmp);
+	if (has_gpu_area(src_pixmap, x, y, width, height))
+		ret = prepare_blt_copy(sna, tmp, alpha_fixup);
+	else if (has_cpu_area(src_pixmap, x, y, width, height))
+		ret = prepare_blt_put(sna, tmp, alpha_fixup);
+	else if (sna_pixmap_move_to_gpu(src_pixmap, MOVE_READ))
+		ret = prepare_blt_copy(sna, tmp, alpha_fixup);
 	else
-		ret = prepare_blt_put(sna, tmp);
+		ret = prepare_blt_put(sna, tmp, alpha_fixup);
 
 	return ret;
 }
@@ -1405,7 +1746,6 @@ fastcall static void sna_blt_fill_op_boxes(struct sna *sna,
 static void sna_blt_fill_op_done(struct sna *sna,
 				 const struct sna_fill_op *fill)
 {
-	blt_done(sna, &fill->base);
 }
 
 bool sna_blt_fill(struct sna *sna, uint8_t alu,
@@ -1419,7 +1759,7 @@ bool sna_blt_fill(struct sna *sna, uint8_t alu,
 
 	DBG(("%s(alu=%d, pixel=%x, bpp=%d)\n", __FUNCTION__, alu, pixel, bpp));
 
-	if (bo->tiling == I915_TILING_Y) {
+	if (!kgem_bo_can_blt(&sna->kgem, bo)) {
 		DBG(("%s: rejected due to incompatible Y-tiling\n",
 		     __FUNCTION__));
 		return FALSE;
@@ -1451,7 +1791,6 @@ static void sna_blt_copy_op_blt(struct sna *sna,
 static void sna_blt_copy_op_done(struct sna *sna,
 				 const struct sna_copy_op *op)
 {
-	blt_done(sna, &op->base);
 }
 
 static void gen6_blt_copy_op_done(struct sna *sna,
@@ -1470,10 +1809,10 @@ bool sna_blt_copy(struct sna *sna, uint8_t alu,
 	return FALSE;
 #endif
 
-	if (src->tiling == I915_TILING_Y)
+	if (!kgem_bo_can_blt(&sna->kgem, src))
 		return FALSE;
 
-	if (dst->tiling == I915_TILING_Y)
+	if (!kgem_bo_can_blt(&sna->kgem, dst))
 		return FALSE;
 
 	if (!sna_blt_copy_init(sna, &op->base.u.blt,
@@ -1489,13 +1828,14 @@ bool sna_blt_copy(struct sna *sna, uint8_t alu,
 	return TRUE;
 }
 
-static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
+static bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 			     struct kgem_bo *bo, int bpp,
 			     uint32_t color,
 			     const BoxRec *box)
 {
 	struct kgem *kgem = &sna->kgem;
 	uint32_t br13, cmd, *b;
+	bool overwrites;
 
 	DBG(("%s: box=((%d, %d), (%d, %d))\n", __FUNCTION__,
 	     box->x1, box->y1, box->x2, box->y2));
@@ -1511,15 +1851,6 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	}
 	assert(br13 < MAXSHORT);
 
-	if (alu == GXclear)
-		color = 0;
-	else if (alu == GXcopy) {
-		if (color == 0)
-			alu = GXclear;
-		else if (color == -1)
-			alu = GXset;
-	}
-
 	br13 |= fill_ROP[alu] << 16;
 	switch (bpp) {
 	default: assert(0);
@@ -1530,15 +1861,40 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	}
 
 	/* All too frequently one blt completely overwrites the previous */
-	if (kgem->nbatch >= 6 &&
-	    (alu == GXcopy || alu == GXclear || alu == GXset) &&
+	overwrites = alu == GXcopy || alu == GXclear || alu == GXset;
+	if (overwrites && kgem->nbatch >= 6 &&
 	    kgem->batch[kgem->nbatch-6] == cmd &&
 	    *(uint64_t *)&kgem->batch[kgem->nbatch-4] == *(uint64_t *)box &&
 	    kgem->reloc[kgem->nreloc-1].target_handle == bo->handle) {
 		DBG(("%s: replacing last fill\n", __FUNCTION__));
 		kgem->batch[kgem->nbatch-5] = br13;
 		kgem->batch[kgem->nbatch-1] = color;
-		return TRUE;
+		return true;
+	}
+	if (overwrites && kgem->nbatch >= 8 &&
+	    (kgem->batch[kgem->nbatch-8] & 0xffc0000f) == XY_SRC_COPY_BLT_CMD &&
+	    *(uint64_t *)&kgem->batch[kgem->nbatch-6] == *(uint64_t *)box &&
+	    kgem->reloc[kgem->nreloc-2].target_handle == bo->handle) {
+		DBG(("%s: replacing last copy\n", __FUNCTION__));
+		kgem->batch[kgem->nbatch-8] = cmd;
+		kgem->batch[kgem->nbatch-7] = br13;
+		kgem->batch[kgem->nbatch-3] = color;
+		/* Keep the src bo as part of the execlist, just remove
+		 * its relocation entry.
+		 */
+		kgem->nreloc--;
+		kgem->nbatch -= 2;
+		return true;
+	}
+
+	/* If we are currently emitting SCANLINES, keep doing so */
+	if (sna->blt_state.fill_bo == bo->unique_id &&
+	    sna->blt_state.fill_pixel == color &&
+	    (sna->blt_state.fill_alu == alu ||
+	     sna->blt_state.fill_alu == ~alu)) {
+		DBG(("%s: matching last fill, converting to scanlines\n",
+		     __FUNCTION__));
+		return false;
 	}
 
 	kgem_set_mode(kgem, KGEM_BLT);
@@ -1546,6 +1902,7 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	    !kgem_check_reloc(kgem, 1) ||
 	    !kgem_check_bo_fenced(kgem, bo, NULL)) {
 		_kgem_submit(kgem);
+		assert(kgem_check_bo_fenced(&sna->kgem, bo, NULL));
 		_kgem_set_mode(kgem, KGEM_BLT);
 	}
 
@@ -1561,8 +1918,10 @@ static Bool sna_blt_fill_box(struct sna *sna, uint8_t alu,
 	b[5] = color;
 	kgem->nbatch += 6;
 
-	sna->blt_state.fill_bo = 0;
-	return TRUE;
+	sna->blt_state.fill_bo = bo->unique_id;
+	sna->blt_state.fill_pixel = color;
+	sna->blt_state.fill_alu = ~alu;
+	return true;
 }
 
 Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
@@ -1580,22 +1939,11 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 	DBG(("%s (%d, %08x, %d) x %d\n",
 	     __FUNCTION__, bpp, pixel, alu, nbox));
 
-	if (bo->tiling == I915_TILING_Y) {
+	if (!kgem_bo_can_blt(kgem, bo)) {
 		DBG(("%s: fallback -- dst uses Y-tiling\n", __FUNCTION__));
 		return FALSE;
 	}
 
-	if (nbox == 1)
-		return sna_blt_fill_box(sna, alu, bo, bpp, pixel, box);
-
-	br13 = bo->pitch;
-	cmd = XY_SCANLINE_BLT;
-	if (kgem->gen >= 40 && bo->tiling) {
-		cmd |= 1 << 11;
-		br13 >>= 2;
-	}
-	assert(br13 < MAXSHORT);
-
 	if (alu == GXclear)
 		pixel = 0;
 	else if (alu == GXcopy) {
@@ -1605,6 +1953,17 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 			alu = GXset;
 	}
 
+	if (nbox == 1 && sna_blt_fill_box(sna, alu, bo, bpp, pixel, box))
+		return TRUE;
+
+	br13 = bo->pitch;
+	cmd = XY_SCANLINE_BLT;
+	if (kgem->gen >= 40 && bo->tiling) {
+		cmd |= 1 << 11;
+		br13 >>= 2;
+	}
+	assert(br13 < MAXSHORT);
+
 	br13 |= 1<<31 | fill_ROP[alu] << 16;
 	switch (bpp) {
 	default: assert(0);
@@ -1617,10 +1976,11 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 	if (!kgem_check_bo_fenced(kgem, bo, NULL) ||
 	    !kgem_check_batch(kgem, 12)) {
 		_kgem_submit(kgem);
+		assert(kgem_check_bo_fenced(&sna->kgem, bo, NULL));
 		_kgem_set_mode(kgem, KGEM_BLT);
 	}
 
-	if (sna->blt_state.fill_bo != bo->handle ||
+	if (sna->blt_state.fill_bo != bo->unique_id ||
 	    sna->blt_state.fill_pixel != pixel ||
 	    sna->blt_state.fill_alu != alu)
 	{
@@ -1649,7 +2009,7 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 		b[8] = 0;
 		kgem->nbatch += 9;
 
-		sna->blt_state.fill_bo = bo->handle;
+		sna->blt_state.fill_bo = bo->unique_id;
 		sna->blt_state.fill_pixel = pixel;
 		sna->blt_state.fill_alu = alu;
 	}
@@ -1674,12 +2034,12 @@ Bool sna_blt_fill_boxes(struct sna *sna, uint8_t alu,
 
 			assert(box->x1 >= 0);
 			assert(box->y1 >= 0);
-			assert(box->y2 * bo->pitch <= bo->size);
+			assert(box->y2 * bo->pitch <= kgem_bo_size(bo));
 
 			b = kgem->batch + kgem->nbatch;
 			kgem->nbatch += 3;
 			b[0] = cmd;
-			*(uint64_t *)(b+1) = *(uint64_t *)box;
+			*(uint64_t *)(b+1) = *(const uint64_t *)box;
 			box++;
 		} while (--nbox_this_time);
 
@@ -1729,8 +2089,13 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 	    src_bo->tiling, dst_bo->tiling,
 	    src_bo->pitch, dst_bo->pitch));
 
-	if (src_bo->tiling == I915_TILING_Y || dst_bo->tiling == I915_TILING_Y)
+	if (!kgem_bo_can_blt(kgem, src_bo) || !kgem_bo_can_blt(kgem, dst_bo)) {
+		DBG(("%s: cannot blt to src? %d or dst? %d\n",
+		     __FUNCTION__,
+		     kgem_bo_can_blt(kgem, src_bo),
+		     kgem_bo_can_blt(kgem, dst_bo)));
 		return FALSE;
+	}
 
 	cmd = XY_SRC_COPY_BLT_CMD;
 	if (bpp == 32)
@@ -1741,7 +2106,7 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 		cmd |= BLT_SRC_TILED;
 		src_pitch >>= 2;
 	}
-	assert(src_pitch < MAXSHORT);
+	assert(src_pitch <= MAXSHORT);
 
 	br13 = dst_bo->pitch;
 	if (kgem->gen >= 40 && dst_bo->tiling) {
@@ -1775,6 +2140,11 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 	    !kgem_check_reloc(kgem, 2) ||
 	    !kgem_check_bo_fenced(kgem, dst_bo, src_bo, NULL)) {
 		_kgem_submit(kgem);
+		if (!kgem_check_bo_fenced(kgem, dst_bo, src_bo, NULL))
+			return sna_tiling_blt_copy_boxes(sna, alu,
+							 src_bo, src_dx, src_dy,
+							 dst_bo, dst_dx, dst_dy,
+							 bpp, box, nbox);
 		_kgem_set_mode(kgem, KGEM_BLT);
 	}
 
@@ -1838,3 +2208,120 @@ Bool sna_blt_copy_boxes(struct sna *sna, uint8_t alu,
 
 	return TRUE;
 }
+
+static void box_extents(const BoxRec *box, int n, BoxRec *extents)
+{
+	*extents = *box;
+	while (--n) {
+		box++;
+		if (box->x1 < extents->x1)
+			extents->x1 = box->x1;
+		if (box->y1 < extents->y1)
+			extents->y1 = box->y1;
+
+		if (box->x2 > extents->x2)
+			extents->x2 = box->x2;
+		if (box->y2 > extents->y2)
+			extents->y2 = box->y2;
+	}
+}
+
+Bool sna_blt_copy_boxes_fallback(struct sna *sna, uint8_t alu,
+				 PixmapPtr src, struct kgem_bo *src_bo, int16_t src_dx, int16_t src_dy,
+				 PixmapPtr dst, struct kgem_bo *dst_bo, int16_t dst_dx, int16_t dst_dy,
+				 const BoxRec *box, int nbox)
+{
+	struct kgem_bo *free_bo = NULL;
+	Bool ret;
+
+	DBG(("%s: alu=%d, n=%d\n", __FUNCTION__, alu, nbox));
+
+	if (!sna_blt_compare_depth(&src->drawable, &dst->drawable)) {
+		DBG(("%s: mismatching depths %d -> %d\n",
+		     __FUNCTION__, src->drawable.depth, dst->drawable.depth));
+		return FALSE;
+	}
+
+	if (src_bo == dst_bo) {
+		DBG(("%s: dst == src\n", __FUNCTION__));
+
+		if (src_bo->tiling == I915_TILING_Y &&
+		    kgem_bo_blt_pitch_is_ok(&sna->kgem, src_bo)) {
+			struct kgem_bo *bo;
+
+			DBG(("%s: src is Y-tiled\n", __FUNCTION__));
+
+			assert(src_bo == sna_pixmap(src)->gpu_bo);
+			bo = sna_pixmap_change_tiling(src, I915_TILING_X);
+			if (bo == NULL) {
+				BoxRec extents;
+
+				DBG(("%s: y-tiling conversion failed\n",
+				     __FUNCTION__));
+
+				box_extents(box, nbox, &extents);
+				free_bo = kgem_create_2d(&sna->kgem,
+							 extents.x2 - extents.x1,
+							 extents.y2 - extents.y1,
+							 src->drawable.bitsPerPixel,
+							 I915_TILING_X, 0);
+				if (free_bo == NULL) {
+					DBG(("%s: fallback -- temp allocation failed\n",
+					     __FUNCTION__));
+					return FALSE;
+				}
+
+				if (!sna_blt_copy_boxes(sna, GXcopy,
+							src_bo, src_dx, src_dy,
+							free_bo, -extents.x1, -extents.y1,
+							src->drawable.bitsPerPixel,
+							box, nbox)) {
+					DBG(("%s: fallback -- temp copy failed\n",
+					     __FUNCTION__));
+					kgem_bo_destroy(&sna->kgem, free_bo);
+					return FALSE;
+				}
+
+				src_dx = -extents.x1;
+				src_dy = -extents.y1;
+				src_bo = free_bo;
+			} else
+				dst_bo = src_bo = bo;
+		}
+	} else {
+		if (src_bo->tiling == I915_TILING_Y &&
+		    kgem_bo_blt_pitch_is_ok(&sna->kgem, src_bo)) {
+			DBG(("%s: src is y-tiled\n", __FUNCTION__));
+			assert(src_bo == sna_pixmap(src)->gpu_bo);
+			src_bo = sna_pixmap_change_tiling(src, I915_TILING_X);
+			if (src_bo == NULL) {
+				DBG(("%s: fallback -- src y-tiling conversion failed\n",
+				     __FUNCTION__));
+				return FALSE;
+			}
+		}
+
+		if (dst_bo->tiling == I915_TILING_Y &&
+		    kgem_bo_blt_pitch_is_ok(&sna->kgem, dst_bo)) {
+			DBG(("%s: dst is y-tiled\n", __FUNCTION__));
+			assert(dst_bo == sna_pixmap(dst)->gpu_bo);
+			dst_bo = sna_pixmap_change_tiling(dst, I915_TILING_X);
+			if (dst_bo == NULL) {
+				DBG(("%s: fallback -- dst y-tiling conversion failed\n",
+				     __FUNCTION__));
+				return FALSE;
+			}
+		}
+	}
+
+	ret =  sna_blt_copy_boxes(sna, alu,
+				  src_bo, src_dx, src_dy,
+				  dst_bo, dst_dx, dst_dy,
+				  dst->drawable.bitsPerPixel,
+				  box, nbox);
+
+	if (free_bo)
+		kgem_bo_destroy(&sna->kgem, free_bo);
+
+	return ret;
+}
