@@ -43,12 +43,32 @@
 
 #define BOUND(v)	(INT16) ((v) < MINSHORT ? MINSHORT : (v) > MAXSHORT ? MAXSHORT : (v))
 
+Bool sna_composite_create(struct sna *sna)
+{
+	xRenderColor color ={ 0 };
+	int error;
+
+	sna->clear = CreateSolidPicture(0, &color, &error);
+	return sna->clear != NULL;
+}
+
+void sna_composite_close(struct sna *sna)
+{
+	FreePicture(sna->clear, 0);
+}
+
 static inline bool
 region_is_singular(pixman_region16_t *region)
 {
 	return region->data == NULL;
 }
 
+static inline bool
+region_is_empty(pixman_region16_t *region)
+{
+	return region->data && region->data->numRects == 0;
+}
+
 static inline pixman_bool_t
 clip_to_dst(pixman_region16_t *region,
 	    pixman_region16_t *clip,
@@ -84,7 +104,7 @@ clip_to_dst(pixman_region16_t *region,
 		}
 
 		return TRUE;
-	} else if (!pixman_region_not_empty(clip)) {
+	} else if (region_is_empty(clip)) {
 		return FALSE;
 	} else {
 		if (dx | dy)
@@ -94,7 +114,7 @@ clip_to_dst(pixman_region16_t *region,
 		if (dx | dy)
 			pixman_region_translate(region, dx, dy);
 
-		return pixman_region_not_empty(region);
+		return !region_is_empty(region);
 	}
 }
 
@@ -116,7 +136,7 @@ clip_to_src(RegionPtr region, PicturePtr p, int dx, int	 dy)
 				-(p->clipOrigin.x + dx),
 				-(p->clipOrigin.y + dy));
 
-	return result && pixman_region_not_empty(region);
+	return result && !region_is_empty(region);
 }
 
 Bool
@@ -231,7 +251,7 @@ sna_compute_composite_region(RegionPtr region,
 		     region->extents.x2, region->extents.y2));
 	}
 
-	return pixman_region_not_empty(region);
+	return !region_is_empty(region);
 }
 
 static void
@@ -395,6 +415,7 @@ sna_composite(CARD8 op,
 {
 	struct sna *sna = to_sna_from_drawable(dst->pDrawable);
 	struct sna_composite_op tmp;
+	unsigned flags;
 	RegionRec region;
 	int dx, dy;
 
@@ -405,6 +426,17 @@ sna_composite(CARD8 op,
 	     dst_x, dst_y, dst->pDrawable->x, dst->pDrawable->y,
 	     width, height));
 
+	if (region_is_empty(dst->pCompositeClip)) {
+		DBG(("%s: empty clip, skipping\n", __FUNCTION__));
+		return;
+	}
+
+	if (op == PictOpClear) {
+		DBG(("%s: discarding source and mask for clear\n", __FUNCTION__));
+		mask = NULL;
+		src = sna->clear;
+	}
+
 	if (mask && sna_composite_mask_is_opaque(mask)) {
 		DBG(("%s: removing opaque %smask\n",
 		     __FUNCTION__,
@@ -425,7 +457,7 @@ sna_composite(CARD8 op,
 		goto fallback;
 	}
 
-	if (dst->alphaMap || src->alphaMap || (mask && mask->alphaMap)) {
+	if (dst->alphaMap) {
 		DBG(("%s: fallback due to unhandled alpha-map\n", __FUNCTION__));
 		goto fallback;
 	}
@@ -461,14 +493,16 @@ sna_composite(CARD8 op,
 		goto fallback;
 	}
 
-	tmp.boxes(sna, &tmp,
-		  REGION_RECTS(&region),
-		  REGION_NUM_RECTS(&region));
+	if (region.data == NULL)
+		tmp.box(sna, &tmp, &region.extents);
+	else
+		tmp.boxes(sna, &tmp,
+			  REGION_BOXPTR(&region),
+			  REGION_NUM_RECTS(&region));
 	apply_damage(&tmp, &region);
 	tmp.done(sna, &tmp);
 
-	REGION_UNINIT(NULL, &region);
-	return;
+	goto out;
 
 fallback:
 	DBG(("%s -- fallback dst=(%d, %d)+(%d, %d), size=(%d, %d)\n",
@@ -476,19 +510,34 @@ fallback:
 	     dst_x, dst_y,
 	     dst->pDrawable->x, dst->pDrawable->y,
 	     width, height));
-
-	sna_drawable_move_region_to_cpu(dst->pDrawable, &region, true);
-	if (dst->alphaMap)
-		sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, true);
+	if (op == PictOpSrc || op == PictOpClear)
+		flags = MOVE_WRITE;
+	else
+		flags = MOVE_WRITE | MOVE_READ;
+	if (!sna_drawable_move_region_to_cpu(dst->pDrawable, &region, flags))
+		goto out;
+	if (dst->alphaMap &&
+	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, flags))
+		goto out;
 	if (src->pDrawable) {
-		sna_drawable_move_to_cpu(src->pDrawable, false);
-		if (src->alphaMap)
-			sna_drawable_move_to_cpu(src->alphaMap->pDrawable, false);
+		if (!sna_drawable_move_to_cpu(src->pDrawable,
+					      MOVE_READ))
+			goto out;
+
+		if (src->alphaMap &&
+		    !sna_drawable_move_to_cpu(src->alphaMap->pDrawable,
+					      MOVE_READ))
+			goto out;
 	}
 	if (mask && mask->pDrawable) {
-		sna_drawable_move_to_cpu(mask->pDrawable, false);
-		if (mask->alphaMap)
-			sna_drawable_move_to_cpu(mask->alphaMap->pDrawable, false);
+		if (!sna_drawable_move_to_cpu(mask->pDrawable,
+					      MOVE_READ))
+			goto out;
+
+		if (mask->alphaMap &&
+		    !sna_drawable_move_to_cpu(mask->alphaMap->pDrawable,
+					      MOVE_READ))
+			goto out;
 	}
 
 	DBG(("%s: fallback -- fbComposite\n", __FUNCTION__));
@@ -497,6 +546,8 @@ fallback:
 		    mask_x, mask_y,
 		    dst_x, dst_y,
 		    width, height);
+out:
+	REGION_UNINIT(NULL, &region);
 }
 
 static int16_t bound(int16_t a, uint16_t b)
@@ -594,6 +645,11 @@ sna_composite_rectangles(CARD8		 op,
 	if (!num_rects)
 		return;
 
+	if (region_is_empty(dst->pCompositeClip)) {
+		DBG(("%s: empty clip, skipping\n", __FUNCTION__));
+		return;
+	}
+
 	if (color->alpha <= 0x00ff) {
 		switch (op) {
 		case PictOpOver:
@@ -631,11 +687,6 @@ sna_composite_rectangles(CARD8		 op,
 	}
 	DBG(("%s: converted to op %d\n", __FUNCTION__, op));
 
-	if (!pixman_region_not_empty(dst->pCompositeClip)) {
-		DBG(("%s: empty clip, skipping\n", __FUNCTION__));
-		return;
-	}
-
 	if (!_pixman_region_init_clipped_rectangles(&region,
 						    num_rects, rects,
 						    dst->pDrawable->x, dst->pDrawable->y,
@@ -652,7 +703,7 @@ sna_composite_rectangles(CARD8		 op,
 	     RegionNumRects(&region)));
 
 	if (!pixman_region_intersect(&region, &region, dst->pCompositeClip) ||
-	    !pixman_region_not_empty(&region)) {
+	    region_is_empty(&region)) {
 		DBG(("%s: zero-intersection between rectangles and clip\n",
 		     __FUNCTION__));
 		pixman_region_fini(&region);
@@ -693,13 +744,13 @@ sna_composite_rectangles(CARD8		 op,
 	 * operation, then we may as well delete it without moving it
 	 * first to the GPU.
 	 */
-	if (op == PictOpSrc || op == PictOpClear) {
+	if (op <= PictOpSrc) {
 		priv = sna_pixmap_attach(pixmap);
-		if (priv && !priv->gpu_only)
+		if (priv)
 			sna_damage_subtract(&priv->cpu_damage, &region);
 	}
 
-	priv = sna_pixmap_move_to_gpu(pixmap);
+	priv = sna_pixmap_move_to_gpu(pixmap, MOVE_READ | MOVE_WRITE);
 	if (priv == NULL) {
 		DBG(("%s: fallback due to no GPU bo\n", __FUNCTION__));
 		goto fallback;
@@ -712,21 +763,66 @@ sna_composite_rectangles(CARD8		 op,
 		goto fallback;
 	}
 
-	if (!priv->gpu_only) {
+	/* Clearing a pixmap after creation is a common operation, so take
+	 * advantage and reduce further damage operations.
+	 */
+	if (!DAMAGE_IS_ALL(priv->gpu_damage)) {
 		assert_pixmap_contains_box(pixmap, RegionExtents(&region));
-		sna_damage_add(&priv->gpu_damage, &region);
+
+		if (region.data == NULL &&
+		    region.extents.x2 - region.extents.x1 == pixmap->drawable.width &&
+		    region.extents.y2 - region.extents.y1 == pixmap->drawable.height) {
+			sna_damage_all(&priv->gpu_damage,
+				       pixmap->drawable.width, pixmap->drawable.height);
+			priv->undamaged = false;
+			if (op <= PictOpSrc) {
+				priv->clear = true;
+				priv->clear_color = 0;
+				if (op == PictOpSrc)
+					sna_get_pixel_from_rgba(&priv->clear_color,
+								color->red,
+								color->green,
+								color->blue,
+								color->alpha,
+								dst->format);
+				DBG(("%s: marking clear [%08x]\n",
+				     __FUNCTION__, priv->clear_color));
+			}
+		} else
+			sna_damage_add(&priv->gpu_damage, &region);
+	} else if (op <= PictOpSrc &&
+		   region.data == NULL &&
+		   region.extents.x2 - region.extents.x1 == pixmap->drawable.width &&
+		   region.extents.y2 - region.extents.y1 == pixmap->drawable.height) {
+		priv->clear = true;
+		priv->clear_color = 0;
+		if (op == PictOpSrc)
+			sna_get_pixel_from_rgba(&priv->clear_color,
+						color->red,
+						color->green,
+						color->blue,
+						color->alpha,
+						dst->format);
+		DBG(("%s: marking clear [%08x]\n",
+		     __FUNCTION__, priv->clear_color));
 	}
 
 	goto done;
 
 fallback:
 	DBG(("%s: fallback\n", __FUNCTION__));
-	sna_drawable_move_region_to_cpu(&pixmap->drawable, &region, true);
-	if (dst->alphaMap)
-		sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, true);
+	if (op <= PictOpSrc)
+		error = MOVE_WRITE;
+	else
+		error = MOVE_WRITE | MOVE_READ;
+	if (!sna_drawable_move_region_to_cpu(&pixmap->drawable, &region, error))
+		goto done;
 
-	if (op == PictOpSrc || op == PictOpClear) {
-		PixmapPtr pixmap = get_drawable_pixmap(dst->pDrawable);
+	if (dst->alphaMap &&
+	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, error))
+		goto done;
+
+	if (op <= PictOpSrc) {
 		int nbox = REGION_NUM_RECTS(&region);
 		BoxPtr box = REGION_RECTS(&region);
 		uint32_t pixel;
